/*-----------------------------------------------------------------------------
  File    : knn.c
  Contents: header file for k-nearest neighbor implementation
  Author  : Frank Rügheimer <frank@ruegheimer.org>
  License : Copyright (C) 2010 Frank Rügheimer
            distributed under the GNU General Public License
            (see doc/license file for terms and conditions)
  History : 2010-02-09 file created (just the scaffold)
            2010-02-10 finished basic command line interface and error handling
            2010-03-02 implemented table i/o, starting work on algorithm
            2010-03-04 moved error codes and column mapping to knn.h
                       implemented basic column mapping
            2010-03-05 program now uses domain file - adapted interface
                       integer inputs now automatically converted to reals
            2010-03-08 using separate attribute sets for each table + general 
                       attribute set now, 
            2010-03-09 working implementation of crude neighbor selection 
            2010-03-10 finished implem./testing equivalence class-based sel.
                       added output attribute to input/output table
            2010-03-11 implemented standard group tie resolution strategy
                       (random resolution not yet implemented)
            2010-03-12 fixed distance check bug during testing
                       1st complete version added -k -c -R options
            2010-03-15 added other fancy options - went to version 0.8
            2010-03-17 changed default target attribute to last column (as it
                       was supposed to be to start with)
            2010-03-18 added -m option for distance measure selection interface
                       and covariance-based distance function
            2010-03-22 added -x options and extensions for cooperation with 
                       Pierre N. (not official part of the interfaces yet)
            2010-03-23 added matching pair exclusion for fast leave-one-out
                       cross-validation. 
            2010-06-08 added "correl2" distance measure; expanded documentation
            2010-06-09 fixed some typos in the help page
                       help now displayed if program is run w/o arguments
                       changed meaning of -c option, introduced -l option
            2010-06-10 implemented -c option and expanded error messages
            2010-07-08 minor corrections on help page -> version 1.01
            2010-09-14 minor corrections of typos in comments and help page
                       -> version 1.02
            2010-09-29 updated release informations + makefile ->version 1.03
            
-----------------------------------------------------------------------------*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#ifndef AS_RDWR
#define AS_RDWR
#endif
#ifndef TAB_RDWR
#define TAB_RDWR
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#include <io.h>
#include <table.h>
#include <time.h>
#include <string.h>
#include "knn.h"

/*----------------------------------------------------------------------------
  Preprocessor Definitions
-----------------------------------------------------------------------------*/
#define PRGNAME     "knn"
#define DESCRIPTION "k-nearest neighbor classifier     "
#define VERSION     "  version 1.03  (2010-09-28)      "\
                    "  (c) 2010      Frank Rügheimer\n"\
                    "  table utility/library package   "\
                    "  (c) 1995-2010 Christian Borgelt\n"\
                    "  used under the terms of the GNU LESSER GENERAL PUBLIC "\
                    "LICENSE\n"
#define COMPDATE    "  compiled on %s\n",__DATE__
/* version number must stay in position 5 as it is parsed by make */
 
                    
/*-----------------------------------------------------------------------------
  Global Variables
------------------------------------------------------------------------------*/
static int      kparam    = 5;          /* neighborhood size for knn (dflt 5) */
const  char     *prgname  = NULL;       /* program name for error messages */
static FILE     *exfile   = NULL;       /* example file (classifier) */
static FILE     *infile   = NULL;       /* input  file */
static FILE     *outfile  = NULL;       /* output file */
static ATTSET   *attset   = NULL;       /* using shared attribute set now */
static ATTSET   *attset_e = NULL;       /* attribute set for the exam table */
static ATTSET   *attset_n = NULL;  			/* attrib. set for tab with new cases */
static TABLE    *table_e  = NULL;       /* table with examples */
static TABLE    *table_n  = NULL;       /* table with new cases */
static COLMAP   *col_etoi = NULL;       /* maps columns indices between files */
static SCAN*    scan      = NULL;       /* scanner for domain file */
static char*    oaname    = "knn";      /* name for output attribute column  */
static char*    cfname    = NULL;       /* name for class att. in input tbls */
static int      resmode   = KNN_NORESOLVE;    /* no tie resolution by default */
/* pointer to currently selected distance function */
  

/*-----------------------------------------------------------------------------
  Functions
-----------------------------------------------------------------------------*/

/* free static memory objects and close files */
void cleanup(){
  if (table_n)  tab_delete(table_n, 0); table_n=NULL; 
  if (table_e)  tab_delete(table_e, 0); table_e=NULL;
  if (attset_n) as_delete(attset_n); attset_n=NULL;
  if (attset_e) as_delete(attset_e); attset_e=NULL; 
  if (attset) as_delete(attset); attset=NULL;
  if ((infile) && (infile != stdin)) fclose(infile); infile=NULL;
  if ((exfile) && (exfile != stdin)) fclose(exfile); exfile=NULL;
  if ((outfile) && (outfile != stdout)) fclose(outfile); outfile=NULL;
  if (col_etoi) cm_delete(col_etoi); col_etoi=NULL;
  if (scan) sc_delete(scan); scan=NULL;
}


/*---------------------------------------------------------------------------*/

/* variable argument error function (seen in tnorm source by CB) */
/* I do mot like the thing with the negative error codes though  */
/* still this is required for compatibility with attset          */
static void error (int code, ...)
{                               /* --- print error message */
  va_list    args;              /* list of variable arguments */
  const char *msg;              /* error message */


  assert(prgname);              /* check the program name */
  if (code < E_UNKNOWN) code = E_UNKNOWN;
  if (code < 0) {               /* if to report an error, */
    msg = errmsgs[-code];       /* get the error message */
    if (!msg) msg = errmsgs[-E_UNKNOWN];
     fprintf(stderr, "\n%s: ", prgname);
    va_start(args, code);       /* get variable arguments */
    vfprintf(stderr, msg, args);/* print the error message */
    va_end(args);               /* end argument evaluation */
  }
  #ifndef NDEBUG                   
  cleanup();                    /* clean up memory */
  #endif
  #ifdef STORAGE
  showmem("at end of program"); /* check memory usage */
  #endif
  exit(code);                   /* abort the program */
}  /* error() */

/*---------------------------------------------------------------------------*/

int showhelp(){                        /* print a help screen */
    assert(prgname);
    printf("USAGE: %s [options] domfile [-h hdrx] xmplfile [-h hdri] infile "\
           "[outfile]\n", prgname);
    printf("OPTIONS:\n");
    printf("  -k#      set neighborhood size parameters (the actual number\n"\
           "           of neighbors may exceed the value as examples having\n"\
           "           equal distance from the current case are included\n"\
           "           as a group)        (default: \"%i\")\n", kparam);
   
    printf("  -R meth  select method for forced tie resolution (initially\n"\
           "           ties are resolved by iteratively removing the most\n"\
           "           distant points from the neighborhood. If this fails,\n"\
           "           no class is assigned. However, tie resolution can be\n"\
           "           enforced by randomly selecting among the candidate\n"\
           "           classes or using the class encountered first)\n"\
           "           valid values: \"random\" ,\"order\"\n"\
           "                              (default: do not resolve)\n"); 
    printf("  -a       align fields       (default: do not align)\n");
    printf("  -w       do not write field names to output file\n");
    printf("  -b#      blank   characters (default: \" \\t\\r\")\n");
    printf("  -f#      field   separators (default: \" \\t\")\n"\
           "  -r#      record  separators (default: \"\\n\")\n");
    printf("  -s       suppress matching of corresponding tupel pairs in  \n"\
           "           example and input files. Used for leave-one-out   \n"\
           "           cross-validation   (default: disabled)\n");
    printf("  -c name  select target attribute column in example file \n"\
           "                              (default:last column)\n");
    printf("  -l name  set name for column classifier output column \n"\
           "                              (default:knn)\n");
    printf("  -m#      select distance/similarity measure\n"\
           "           permissible values are: \n"\
           "           \"euclid \":      ordering with euclidean distance\n"\
           "           \"correl \":      ordering based on correlation \n"\
           "                           neg. correlation increases distance\n"\
           "           \"correl2\":      ordering based on correlation \n"\
           "                           neg. correlation decreases distance\n"\
           "                              (default: euclidean distance)\n");
    printf("  -C#      comment characters (default: \"#\")\n");
    printf("  -x       extended output \n"\
           "           Setting this option adds columns with average value\n"\
           "           of similarity/distance to members of neighborhood,\n"\
           "           number of occurrences for the most frequent class in\n"\
           "           the neighborhood and the real size of the selected\n");
    printf("           neighborhood. The latter may differ from the argument\n"\
           "           passed via -k due to the extended definition the\n"\
           "           neighborhood and the tie resolution strategy\n"\
           "           Note: distances will currently be averaged even if \n"\
           "           that operation is not meaningful for the selected \n"\
           "           measure. This may change in future versions\n");
    printf("  -h       read table header (field names) from hdrfile\n"\
           "                    (default: field names in first record)\n");
    printf("  hdrx     file containing table header (field names) for \n"\
           "           example file\n");
    printf("  hdri     file containing table header (field names) for \n"\
           "           input case file\n");
    printf("ARGUMENTS:\n");
    printf("  domfile  file with domain specifications (see dom utility)\n");
    printf("  xmplfile file containing classification examples (no header)\n");
    printf("  infile   file containing cases to be classified (no header)\n");
    printf("  outfile  output file for table with classification results to\n");
    printf("           if omitted output is written to <stdout>\n");   
    return 0;                  
}

/*=============================================================================
   functions specific to the knn implementation 
   rest is a general classifier plus knn-specific command-line parameters
-----------------------------------------------------------------------------*/

/* compute the distance between two tuples */
/* assumes only real-valued attributes are enabled */
double dist_euclid(const TUPLE* datatpl,const TUPLE* examtpl,const COLMAP* cm){
   int    i,j;   /* current attribute indices */       
   double d;     /* distance of in values for current attribute */
   double sqsum; /* to sum squared distances over attributes */
   int    n;     /* number of tuple pairs considered */
   sqsum=0.0;
   for (i=n=0;i<tpl_colcnt(examtpl); i++){
     j=cm_mapidx(cm,i);
     if (j>=0) {
       d=*(float*)(tpl_colval(examtpl,i))   /* compute distance between ... */
        -*(float*)(tpl_colval(datatpl,j));  /* ... instantiations           */
       sqsum+=(d*d);                        /* increase aggreg. sqared sum  */
       n++;                                 /* and count as valid attribute */
     }
   }
   if (n==0) error(E_TAB_ATTS);
   /* since we are just sorting there is no need to actually compute the root*/
   return(sqsum);                                                                    
}

/* FR: untested  continue here this has to be finished */
/* value is 0   for fully correlated, 
            0.5 for uncorrelated and 
            1.0 for maximal inverse  correllation */
double dist_corr(const TUPLE* datatpl, const TUPLE* examtpl,const COLMAP* cm){ 
   int    i,j;             /* current attribute indices to trav. tupels */       
   double d_x,d_y;         /* distance to Expected value */
   double mu_x,mu_y;       /* means  */
   double d_xx,d_yy;       /* to aggregate empirical variances */
   double d_xy;            /* covariance  */
   int    n;               /* number of value pairs considered */
   double corr;            /* correlation coefficient */
   mu_x=mu_y=d_xx=d_yy=d_xy=0.0;
   for (i=n=0;i<tpl_colcnt(examtpl); i++){     /* 1st traversal */
     j=cm_mapidx(cm,i);
     if (j>=0) {
       mu_x+=*(float*)(tpl_colval(examtpl,i)); /* aggreg values from tpl 1 */
       mu_y+=*(float*)(tpl_colval(datatpl,j)); /* aggreg values from tpl 2 */
       n++;                                    /* and count as valid attrib. */
     }
   }
   if (n==0) error(E_TAB_ATTS); else{
     mu_x/=n; mu_y/=n;                         /* compute means */
     for (i=0;i<tpl_colcnt(examtpl); i++){     /* 2nd traversal */
       j=cm_mapidx(cm,i);
       if (j>=0) {
         d_x=(*(float*)tpl_colval(examtpl,i))-mu_x;
         d_y=(*(float*)tpl_colval(datatpl,j))-mu_y;
         d_xy+=(d_x*d_y); 
         d_xx+=(d_x*d_x); d_yy+=(d_y*d_y);
       }
     }
   }
   /* give worst scores to constant genes */
   if ((d_xx==0.0)||(d_yy==0.0)) return 1;  /* if undef. use maximum value */
   corr=(n<=1)?1:(d_xy)/sqrt(d_xx*d_yy);
   /* since we are just sorting there is no need to actually compute the root
      however we do compute it to support mean distance and other measures  */
   return(0.5*(1-corr));                                                                    
}
/*---------------------------------------------------------------------------*/

/* value is 0   for fully correlated, 
            0   for negatively correlated and 
            1.0 uncorrelated */
double dist_corr2(const TUPLE* datatpl, const TUPLE* examtpl,const COLMAP* cm){ 
   int    i,j;             /* current attribute indices to trav. tupels */       
   double d_x,d_y;         /* distance to Expected value */
   double mu_x,mu_y;       /* means  */
   double d_xx,d_yy;     /* to aggregate empirical variances */
   double d_xy;          /* covariance  */
   int    n;     /* number of value pairs considered */
   double corr;            /* correlation coefficient */
   mu_x=mu_y=d_xx=d_yy=d_xy=0.0;
   for (i=n=0;i<tpl_colcnt(examtpl); i++){     /* 1st traversal */
     j=cm_mapidx(cm,i);
     if (j>=0) {
       mu_x+=*(float*)(tpl_colval(examtpl,i)); /* aggreg values from tpl 1  */
       mu_y+=*(float*)(tpl_colval(datatpl,j)); /* aggreg valued from tpl 2  */
       n++;                                    /* and count as valid attrib */
     }
   }
   if (n==0) error(E_TAB_ATTS); else{
     mu_x/=n; mu_y/=n;                         /* compute means */
     for (i=0;i<tpl_colcnt(examtpl); i++){     /* 2nd traversal */
       j=cm_mapidx(cm,i);
       if (j>=0) {
         d_x=(*(float*)tpl_colval(examtpl,i))-mu_x;
         d_y=(*(float*)tpl_colval(datatpl,j))-mu_y;
         d_xy+=(d_x*d_y); 
         d_xx+=(d_x*d_x); d_yy+=(d_y*d_y);
       }
     }
   }
   /* give worst scores to constant genes */
   if ((d_xx==0.0)||(d_yy==0.0)) return 1;  /* if undef. use maximum value */
   corr=(n<=1)?1:(d_xy)/sqrt(d_xx*d_yy);
   /* since we are just sorting there is no need to actually compute the root
      however we do compute it to support mean distanced and other measures  */
   return(1-(fabs(corr)));                                                                    
}


/*----------------------------------------------------------------------------*/

/* curently features to real data + nominal target att */
int knn_assign(int k, const TABLE* exam, TABLE* data, TPLDIST *tpldist,
               const COLMAP *cm, int mode){
  NGHREC* nghlist = NULL;       /* list of classes to be initialized */
  TUPLE*  examtpl = NULL;       /* to traverse example tuples */
  TUPLE*  datatpl = NULL;       /* to traverse data tupels */
  int     i,j;                  /* to traverse tupels in data and xmpl files */
  double  d,cdst;               /* distance between tuples of current pairing */
  INST    ccls;                 /* current class associated with tuple */
  int     numngh;               /* number of points curr. considered a nearest 
                                   nghbor in neighborlist (may differ from k)*/ 
  NGHREC* cur;                  /* record currently sel. for el. insertion */
  NGHREC* lastgrp = NULL;       /* start ptr for last group of equidist. pts 
                                   that are still considered nearest neighbors*/
  NGHREC* oth;                  /* used for copying and traversing records */
  int     outcol  = -1;         /* write classif. res. to this column in data*/
  int     clscol  = -1;         /* column number of class attribute */
  int     ninst;                /* number of instantiations for output att. */
  int*    instfrq = NULL;       /* note instantiation freqs. in nghbhd */  
  int*    cref;                 /* for backlinking indices of other instant. */
                                /* that have same frequency */ 
  int     ctop = 0;             /* index of most recently seen max freq. cls */
  int     nmax = 0;             /* number. of max. freq. classes */ 
  int     cidx;                 /* for index of class value */    
  double  avg_dst;              /* for calculating average distance */
              
  
  /* This version of the knn neighbor detection algorithm traverses for each 
     case the list of example, maintaining a list of the currently known nea-
     rest neighbors. If a tie for the k-th nearest neighbor arises it keeps  
     track of the whole equivalence class of equidistant points. Node 
     insertions are conducted by swapping entries. Neighbor list is kept 
     sorted up to lastel*/
 
  /* this may later be set in via the argument list */
  
  ATT*    clsatt  = NULL;
  int     clstype;                          /* type of the class attribute */      
  
  assert(data); assert(k>0);                /* check arguments */
  assert(exam&&tab_tplcnt(exam)>0);         /* we want at least 1 exmpl */
  outcol=cm->outcol;                        /* shortcut */
  if (outcol<0) outcol=tab_colcnt(data)-1;  /* pick outcol if not given */
  /* get target attribute from table */
  clscol=cm->clscol;
  if (clscol<0) clscol=tab_colcnt(exam)-1; /* if no class-col, take last one */ 
  if (!(clsatt=tab_col(exam,clscol))) error(E_NOCLSATT);
  #ifndef NDEBUG
  fprintf(stderr,"class attribute is : %s (no. %i)\n",att_name(clsatt),clscol);
  #endif
  if (!((clstype=att_type(clsatt))==(AT_NOM))) 
    error(E_CATTYPE,att_name(clsatt)); 
  /* allocate memory for data structures */;
  ninst=att_valcnt(clsatt);                      /* get target domain size */
  instfrq=malloc(2*ninst*sizeof(int));            
  nghlist=malloc(tab_tplcnt(exam)*sizeof(NGHREC));
  if ((!nghlist)||(!instfrq)) {
    /* if only one of the above succeeds these lists will be freed on exit */
    error(E_NOMEM);
  }
  cref=instfrq+ninst;  /* sub-array for managing set of max. freq instant. */
  
  for (i=0;i<tab_tplcnt(data);i++) {
    datatpl=tab_tpl(data, i);  /* extract tuple */
    numngh=0;                  /* (re-)init list state */
    lastgrp=nghlist;           
    for (j=0; j<tab_tplcnt(exam); j++) {
       if ((mode & KNNX_SAME) && (i==j)) 
         continue; /* skip corr. tuple pair */
       examtpl=tab_tpl(exam, j);            /* get next example tuple */
       cdst=(*tpldist)(datatpl,examtpl,cm); /* get distance and ...   */
       ccls=*(tpl_colval(examtpl,clscol));  /* extract class code     */     
       /* check if to be inserted and update neighbor list            */    
       /* if to insert value seek the insertion pos and make space    */     
       for (oth=cur=(nghlist+numngh);
            (cur>nghlist)&&(((--oth)->dist)>cdst);cur--){                                
          cur->dist=oth->dist;              /* shift entries to make room */
          cur->class=oth->class;            /* ... for insertion and ...  */
          if (oth==lastgrp) lastgrp++;      /* update lastgrp if required */         
       }  
       if ((cur>=nghlist+k)&&(cdst>lastgrp->dist))
          continue;   /* no neighbor -- skip update */
       cur->dist=cdst; cur->class=ccls;  numngh++; /* note new neighbor data */
       if (cur>lastgrp) { /* insertion after lastgrp */
         if (cdst>lastgrp->dist) {
            lastgrp=cur; /* entry in range and forms new last group */
         }               /* otherwise append it extends the last group */
       }  
       else if ((lastgrp)>=(nghlist+k)){ /* if last group has become invalid */
         /* ... seek new last neighbor group start */
         oth=lastgrp=nghlist+k-1; numngh=k;
         for (d=oth->dist,lastgrp--;((lastgrp->dist)==d)&&(lastgrp>nghlist);
                                                                   lastgrp--);
         lastgrp++;
       }            
#ifndef NDEBUG
       /* TESTING: show sorted list of current neighbors */
       for (oth=nghlist;oth<(nghlist+numngh);oth++){
         fprintf(stderr,"(%.1f, %i), ",oth->dist,(oth->class).i);
       }     
       fprintf(stderr,"  (%i elements)\n\n", numngh); 
#endif
    } 
    /* select output class and write it to target column */
    /* clear instfrq list */
    if (instfrq!=memset(instfrq, 0, ninst*sizeof(int))) /* clear counters */
      error(E_UNKNOWN);        /* memset should never change the pointer! */
    for (oth=nghlist;oth<nghlist+numngh;oth++){
      /* not clean as it uses the knowledge that instances are encoded
         via consecutive 0-based value ids (which can used for array indices)
         check if recompiling with later versions */
      cidx=(oth->class.i); 
      if (cidx<0||cidx>=ninst) error(E_ATTVAL);
      (instfrq[cidx])++;  /* register class of item */ 
    }
    /* find most frequent class */
    while (numngh>0) {  /* unless there are no neighbors left... */ 
      /* fprintf(stderr,"%i\n", numngh); */ /* size of nghb.-set for debug */
      /* find the maximal frequent classes in current neighbor selection   */ 
	  nmax=1; ctop=0; cref[0]=0;
	  for (j=1;j<ninst;j++){            /* traverse instance counters  */
	    if (instfrq[j]>instfrq[ctop]) {
	      nmax=1; ctop=j; cref[j]=j;    /* mark top inst; start new inst set */  
	    }
	    else if (instfrq[j]==instfrq[ctop]) { 
	      nmax++; cref[j]=ctop; ctop=j; /* append to list if maxfrq inst. */ 
	    } 
	  }
	  if (nmax==1){ /* if there is a unique most frequent instantiation... */
	    tpl_colval(datatpl,outcol)->i=ctop; /* assign it to the output var */
	    if (mode & KNNX_DIST) { /* add distance column */
	       /* report average distance of winning class biased area/dist */
	       avg_dst=0; 
	       for (oth=nghlist+numngh;--oth>=nghlist;) { /*accumul. dists 4 class */ 
	         if ((oth->class).i==ctop){               
	           if (tpldist==(&dist_euclid)) avg_dst+=sqrt(oth->dist);
	           else avg_dst+=oth->dist;
	         }
	       }
	       tpl_colval(datatpl,cm->dstcol)->f=(avg_dst)/((float)(instfrq[ctop]));	    
	    } 
	    if (mode & KNNX_NNUM) { /* fill column with neighbor number */
	       tpl_colval(datatpl,cm->nnghcol)->i=numngh;
	    }
	    if (mode & KNNX_MFRQ) { /* report abs. frq. of maxfrq class */
	       tpl_colval(datatpl,cm->mfrqcol)->f=((double)(instfrq[ctop])/numngh);
	    }
	    break;                              /* and break from the loop */
	  }
	  /* otherwise drop the last group and retry ... */
	  oth=(nghlist+numngh-1),d=oth->dist;
	  for (;(oth>=nghlist)&&(oth->dist)>=d;oth--){ /* enforce boundaries */
	     cidx=(oth->class.i); 
         if (cidx<0||cidx>=ninst) error(E_ATTVAL);
	     instfrq[cidx]--;
	     numngh--;
	  }
    }; /* must stop when run out of neighbors */
    if ((nmax>1)&&(examtpl)) /* if still tied  and we have examples at all */
    switch (resmode){        /* ...and resolution method selected */
      case KNN_RESORDER:  /* pick max freq. class fist observed in examples */ 
         *tpl_colval(datatpl,outcol)=nghlist->class; break;
      case KNN_RESRANDOM: /* pick randomly among max frequent classes */
         j=rand()%nmax;
         for (cidx=ctop;--j>=0;cidx=cref[cidx]); /* select a random max cls */
         tpl_colval(datatpl,outcol)->i=cidx; break;
      default: break;   
    }               
  }  
  /* free the dynamic memory */ /* something messed with the freelist */
  if (nghlist) free(nghlist);
  if (instfrq) free(instfrq); 
  fprintf(stderr,"done\n");
  return(0);
}


/*-----------------------------------------------------------------------------
   column mapping functions
-----------------------------------------------------------------------------*/
/* create a column mapping   */
COLMAP* cm_create(TABLE* from, TABLE* to, const char* tcname){ 
  ATT*    att_from;                     /* to traverse column attributes */
  ATT*    att_to;
  int     i,j;                          /* counters */ 
  int     sz_from;                      /* size of domain (# cols in "from") */                      
  int     sz_to;                        /* number of columns in table "to" */
  int     atype;                        /* current attribute type */ 
  COLMAP* cm=NULL;                      /* column map to be created */
  int     foundcls=0;                   /* has the named class been found? */
  
  assert(from && to);                   /* check inputs */
  
  
  cm=malloc(sizeof(COLMAP)); if (!cm) error(E_NOMEM); /* create structure */
  sz_from=(cm->numcols)=tab_colcnt(from);
  sz_to=tab_colcnt(to);
  /* construct structure in memory */
  cm->mapsto=malloc(sz_from*sizeof(int));
  if (!(cm->mapsto)){
    free(cm); cm=NULL; error(E_NOMEM); 
  }
  /* init special column references */
  cm->dstcol    = (-1);           /* column no. for distance */
  cm->mfrqcol   = (-1);           /* column no. for maximal class frq */
  cm->nnghcol   = (-1);           /* column for number of neighbors */
  cm->outcol    = (-1);           /* column in data file to write output to */
  cm->clscol    = (sz_from-1);    /* column in example file with class att. */
  /* get mapping function, set default class column */
  for (i=0; i<sz_from; i++) {          /* build mapping */
    att_from=tab_col(from,i);
    atype=att_type(att_from);
    cm->mapsto[i]=(-1);                /* mark attribute as unmapped */
    if (tcname && strcmp(att_name(att_from),tcname)==0 ){
       cm->clscol=i;                   /* set named target attribute    */
       foundcls=1;                     /* remember successful assignent */
    }
    if (atype!=AT_REAL)  {             /* can handle real valued atts only */
       fprintf(stderr,"attribute \"%s\" is not real-valued - skipping ...\n",
                      att_name(att_from));
       continue;                       /* and skip ahead to next attrib */ 
    }
    for (j=0; j<sz_to; j++) {          /* seek matching attribute in image...*/
      att_to=tab_col(to,j);            /* ..simple comparison -> w/o domains */
      if (atype!=att_type(att_to)) continue;
      if (strcmp(att_name(att_from),att_name(att_to))==0) {
        cm->mapsto[i]=j;
        #ifndef NDEBUG       
        fprintf(stderr,"matching columns: %i (%s) <-> %i (%s) \n",
                  i, att_name(att_from),j,att_name(att_to));
        #endif
        break;
      }            
    }  
  }
  if (tcname && !foundcls)
    fprintf(stderr,"could not find attribute \"%s\" - using default\n",tcname); 
  return(cm);      
}

/*---------------------------------------------------------------------------*/

void cm_delete(COLMAP* cm){
  if (cm) {
    if (cm->mapsto) free(cm->mapsto);
    free(cm);
  }
}

/*---------------------------------------------------------------------------*/

#ifndef NDEBUG
/* used for testing */
void as_show(ATTSET* attset){
  ATT* att;
  int i;
  assert(attset);
  fprintf(stderr,"\n-----------------------------------------------\n");
  fprintf(stderr,  "Info on attribute set: %s\n",as_name(attset));
  fprintf(stderr,  "attribute list:\n");
  for (i=0;i<as_attcnt(attset);i++){
    att=as_att(attset,i);
    fprintf(stderr,"\t%s(%p): type: %i, #values: %i, ,marked: %i\n",
       att_name(att),(void*)att,att_type(att),att_valcnt(att),att_getmark(att)); 
  }
}

void cm_show(COLMAP* cm){}
#endif  

/*=============================================================================
   main function
-----------------------------------------------------------------------------*/

/* TODO: parse table using christian's scanner rather than current method for 
   more robust detection of columns and column types */
int main(int argc, char* argv[]){ 
  char    *s;                    /* to traverse options */
  char    **optarg  = NULL;      /* option argument */
  char    *fn_dom   = NULL;      /* name of domain file */
  char    *fn_exam  = NULL;      /* name of example file */
  char    *fn_in    = NULL;      /* name of table file */
  char    *fn_out   = NULL;      /* name of output file */
  char    *fn_hdr_e = NULL;      /* names of header files ... */
  char    *fn_hdr_n = NULL;      /* (currently not supported) */
  char    *rmetname = NULL;      /* name of the tie resolution method */
  int     i,j,k,d;               /* counters, error code variable */ 
  clock_t t;                     /* timer for run-time measurement (tnx CB) */
  /*int     classcol  = (-1); */     /* index of class column in example file*/
  int     ctype;                 /* type of current selected column */
  int     attcnt;                /* number of attributes in domain file */
  ATT*    tatt;                  /* for adding the output attribute */
  ATT*    xatt;                  /* auxiliary for accessing attribute */
  char    *blanks  = NULL;       /* blanks */
  char    *fldseps = NULL;       /* field  separators */
  char    *recseps = NULL;       /* record separators */
  char    *nullchs = NULL;       /* null value characters */
  char    *comment = NULL;       /* comment characters */  
  char    *dstname = NULL;       /* name of distance function */
  int     flags_e  = AS_MARKED;  /* flags for reading example file ... */ 
  int     flags_n  = AS_MARKED;  /* ... and file cases to classify     */
  int     flags_o  = AS_MARKED|AS_ATT;  /* table write flags for output file */
  TPLDIST* tpldist = &dist_euclid;/* function used to compute tuple distances*/
  int     knnxmode=0;            /* flag for extended columns */
  int     extcols=0;             /* for numbering additional columns */
  
  /* parse command line options and arguments */
  prgname = argv[0];                  /* get program name form command line */
  /* --- print startup/usage message --- */
  if (argc != 0) {                    /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION);         /* print a start-up message */
    fprintf(stderr, COMPDATE );}      /* ... and version info    */
  else {                              /* if no argument is given */
     showhelp();                      /* print help screen...*/
     exit(0);                         /* and abort the program */
  }
  srand (time(NULL));                 /* init random number generator */
  /* parse the command line */
  for (i = 1, k=0; i < argc; i++) {   /* traverse arguments     */
    s = argv[i];                      /* get option argument    */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {        /* -- if argument is an option */
      while (1) {                     /* traverse characters */
        switch (*s++) {               /* evaluate option */
          case 'k': kparam=(int)strtol(s, &s, 0); break; /* set nghbrhd size */
          case 'R': optarg=&rmetname; break;                
          case 'c': optarg=&cfname;   break;
          case 'l': optarg=&oaname;   break;
          case '?': showhelp();exit(2);        break;
          /* these are not documented yet */
          case 'a': flags_o |= AS_ALIGN;              break; /* align o. cols*/
          case 'w': flags_o &= ~AS_ATT;               break; /* no output hdr*/
          case 'b': optarg  =  &blanks;               break;
          case 'f': optarg  =  &fldseps;              break;
          case 'r': optarg  =  &recseps;              break;
          case 's': knnxmode |= KNNX_SAME;            break;
          case 'u': optarg  =  &nullchs;              break;
          case 'C': optarg  =  &comment;              break;
          case 'm': optarg  =  &dstname;              break;
          /* case 'd': flags  |= AS_DFLT;               break;*/
          case 'h': switch (k) {
                      case 1: optarg  = &fn_hdr_e;      break;
                      case 2: optarg  = &fn_hdr_n;      break;
                      default: error(E_HDRSPEC);      break;
                    }                                 break;  
          case 'x': knnxmode |= KNNX_DIST|
                               KNNX_NNUM|
                               KNNX_MFRQ;             break;
          default : error(E_OPTION, *--s);            break;
        }                             /* set option variables */
        if (!*s) break;               /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } 
    }                                 /* get option argument */
    else {                            /* -- if argument is no option */
      switch (k++) {                  /* evaluate non-option */
        case  0: fn_dom   = s;    break;
        case  1: fn_exam  = s;    break;
        case  2: fn_in    = s;    break;
        case  3: fn_out   = s;    break;
        default: error(E_ARGCNT); break;
      }                               /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);             /* check option argument and ... */
  if (k==0) {showhelp();  exit(0); }       /* show help when no args given  */
  if ((k < 3) || (k > 4)) error(E_ARGCNT); /*... the number of arguments read*/
  /* nice trick about specifying stdin as input files (tnx christian):
     if '-' is specified as a command line argument the next free arg 
     will be assigned the empty string - is removed by option 
     processing before argument assignment is conducted               */     
  if (fn_hdr_e && (strcmp(fn_hdr_e, "-") == 0))
    fn_hdr_e = "";                /* convert "-" to "" */    
  if (fn_hdr_n && (strcmp(fn_hdr_n, "-") == 0))
    fn_hdr_n = "";                /* convert "-" to "" */            
  i = (!fn_exam || !*fn_exam) ? 1 : 0;
  if  (!fn_dom || !*fn_dom) i++;
  if  (!fn_in || !*fn_in) i++;      
  if  ( fn_hdr_e && !*fn_hdr_e) i++;        
  if  ( fn_hdr_n && !*fn_hdr_n) i++;  /* check assignments of stdin:  */ 
  if (i > 1) error(E_STDIN);          /* stdin can be used once at most */
  if (kparam <= 0) error(E_OPTARGVAL,"for option k - call with -? for help");
  if (fn_hdr_e)                   /* set the header file flag */
     flags_e = AS_ATT | (flags_e & ~AS_DFLT); 
  if (fn_hdr_n)                   /* set the header file flag */
     flags_n |= AS_ATT | (flags_n & ~AS_DFLT); 
  if ((flags_o & AS_ALIGN) && (flags_o & AS_ATT))
     flags_o |= AS_ALNHDR;        /* set align to header flag */
  if (rmetname) {
    if (strcmp(rmetname,"random")==0) resmode=KNN_RESRANDOM;
    else if (strcmp(rmetname,"order")==0) resmode=KNN_RESORDER;
    else error(E_OPTARGVAL,"for option R - call with -? for help "); 
  } 
  /* set selected distance function */
  if (dstname) {
    if (strcmp(dstname,"euclid")==0) tpldist=&dist_euclid;
    else if (strcmp(dstname,"correl")==0) tpldist=&dist_corr;
    else if (strcmp(dstname,"correl2")==0) tpldist=&dist_corr2;
    else error(E_DISTFUN,dstname);
  }
  fprintf(stderr, "\n");        /* terminate the start-up message */
  /* can still be overridden by getting one of the files from stdin */

  /* read example files */
  t     = clock();                    /* start the timer */
  attset=as_create("domains",att_delete); 
  if (!attset) error(1);
  /* configure separators */
  as_chars(attset, recseps, fldseps, blanks, nullchs, comment);
  /* read example domain specifications from domfile */ 
  scan=sc_create(fn_dom);
  if (!scan) error((!fn_dom || !*fn_dom) ? E_NOMEM : E_FOPEN, fn_dom);
  fprintf(stderr, "reading %s ... ", sc_fname(scan));
  if ((sc_nexter(scan)   <  0)  /* start scanning (get first token) */
  ||  (as_parse(attset, scan, AT_ALL) != 0)
  ||  (as_attcnt(attset) <= 0)  /* parse attribute set and */
  ||  !sc_eof(scan))            /* check for end of file */
    error(E_PARSE, sc_fname(scan));
  attcnt = as_attcnt(attset);   /* print a success message */
  fprintf(stderr, "[%d attribute(s)] done ", attcnt);
  fprintf(stderr, "[%.2fs].\n", SEC_SINCE(t));
  sc_delete(scan); scan=NULL; /* cleanup scanner */
  
  /* (christian modifies separators etc. here) */
  attset_e=as_clone(attset);
  attset_n=as_clone(attset);
  if (!attset_e||!attset_n) error(1); 
  as_rename(attset_e,"example data domains");
  as_rename(attset_n,"input data domains"); 
  
  table_e=io_tabin(attset_e,fn_hdr_e, fn_exam, \
                    flags_e,"table examples", 1);
  if (!table_e) error(1);  /* read the example table */
  table_n=io_tabin(attset_n,fn_hdr_n,\
                   fn_in,flags_n,"tab(cases to be classified)",1); 
  if (!table_n) error(1);    /* read the input cases table */
  
#ifndef NDEBUG
  as_show(attset);  as_show(attset_e);  as_show(attset_n);
#endif 
 
 
  fprintf(stderr,"example table has %i columns\n",tab_colcnt(table_e));
  /* convert integer columns to real */
  for (i=tab_colcnt(table_e);(--i)>=0;){
     ctype=tab_coltype(table_e,i);
     switch (ctype){
       case AT_INT: tab_colconv(table_e, i, AT_REAL);
                    fprintf(stderr,"integer attribute %s converted to real\n",
                            att_name(tab_col(table_e,i)));break;
     } 
  }
  for (i=0,j=tab_colcnt(table_n);i<j;i++){
      if (tab_coltype(table_n,i)==AT_INT)
        tab_colconv(table_n, i, AT_REAL);      
  }
  
  /* construct column mapping between files */
  col_etoi=cm_create(table_e,table_n,cfname);       
  if (!col_etoi) error(E_NOMEM);
  /* add output attribute to input attribute set */
  for (i=tab_colcnt(table_n);--i>=0;){       /* to check for name conflicts  */ 
    xatt=tab_col(table_n,i);                 /* ...traverse table attributes */
    if (strcmp(att_name(xatt),oaname)==0){   /* if att of same name is found */
      if (att_getmark(xatt)>=0)              /* ... and is used already ...  */
        error(E_TATTNAME,oaname);            /* fail gracefully ...          */
      else tab_colrem(table_n,i);            /* otherwise - remove the att   */
    }                                    
  } 
  /*  insert result column */ 
  tatt=att_clone(tab_col(table_e,col_etoi->clscol));
                                             /* create target attribute ... */
  att_rename(tatt,oaname);                   /* as copy of class attrib ... */
  att_setmark(tatt,1); 
  tab_coladd(table_n,tatt,1);                /* and insert it to in/out tbl */
  
  /* make extra columns and insert them into table */
  if  (knnxmode) {
    extcols=col_etoi->numcols;
    col_etoi->outcol=extcols++;
    if (knnxmode & KNNX_DIST){
      tatt=att_create("avg_dist",AT_REAL);
      att_setmark(tatt,1);
      if (tab_coladd(table_n,tatt,1)) error(E_TCOLADD,att_name(tatt));
      col_etoi->dstcol=extcols++;
    }
    if (knnxmode & KNNX_MFRQ){
      tatt=att_create("cmaxfrq",AT_REAL);
      att_setmark(tatt,1);
      if (tab_coladd(table_n,tatt,1)) error(E_TCOLADD,att_name(tatt));
      col_etoi->mfrqcol=extcols++;
    }
    if (knnxmode &KNNX_NNUM){ 
      tatt=att_create("#nghb. cons.",AT_INT);
      att_setmark(tatt,1);
      if (tab_coladd(table_n,tatt,1)) error(E_TCOLADD,att_name(tatt));
      col_etoi->nnghcol=extcols++;
    }
  }
  
  if ((d=knn_assign(kparam,table_e,table_n,tpldist,col_etoi,knnxmode))) 
    error(d);                          /* run knn and capture error messages */
  
 
 
    
  /* write (modified input file to output), report errors if necessary */
  if ( (d=io_tabout(table_n,fn_out,flags_o,1)) )    
    error(1); 
  fprintf(stderr, "done [%.2fs].\n", SEC_SINCE(t));
  cleanup();
  return 0;                   /* indicate success */
}
