/*------------------------------------------------------------------------------
  File    : knn.h
  Contents: header file for k-nearest neighbor implementation
  Author  : Frank Rügheimer
  License : GNU General Public License (once published)
  TODO    : check on license version to use (christian's stuff was under 2.x) 
            may I go to 3.x without breaking that license (should be fine for 
            the libs I used) 
  History : 2010-02-24 file created
            2010-03-02 turns out this is pretty useless - will probably delete 
                       it later on
            2010-02-04 added column mapping data structure      
            2010-03-15 error flags adapted for version 0.8      
            2010-03-22 expanded COLMAP specification and added ext. mode flags 
            2010-06-10 changed error messages and expanded column map definition
------------------------------------------------------------------------------*/
#ifndef __KNN__
#define __KNN__
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#ifndef AS_RDWR
#define AS_RDWR
#endif 
#ifndef TAB_RDWR
#define TAB_RDWR
#endif
#include <io.h>
#include <table.h>

/* timer (also from christian) */                  
#define SEC_SINCE(t)  ((clock()-(t)) /(double)CLOCKS_PER_SEC)

/* tie resolution mode flags */
#define KNN_NORESOLVE  0         /* no tie resolution - leave unassigned   */
#define KNN_RESORDER   1         /* select class that occurs first in exam */
#define KNN_RESRANDOM  2         /* randomly select between tied classes   */


/* extension mode flags */
#define KNNX_NONE   0x0000             /* all extensions disabled */
#define KNNX_DIST   0x0001             /* enable distance column */
#define KNNX_NNUM   0x0002             /* enable actual neighbor count column */
#define KNNX_MFRQ   0x0004             /* enable maximal class frq. column */ 
#define KNNX_SAME   0x0008             /* skip corresponding tupel pos.       */
                            
  
/* --- error codes --- */      
#define OK             0         /* no error */
#define E_NONE         0         /* no error */
#define E_NOMEM      (-1)        /* not enough memory */
#define E_FOPEN      (-2)        /* file open failed */
#define E_FREAD      (-3)        /* file read failed */
#define E_FWRITE     (-4)        /* file write failed */
#define E_OPTION     (-5)        /* unknown option */
#define E_OPTARG     (-6)        /* missing option argument */
#define E_ARGCNT     (-7)        /* wrong number of arguments */
#define E_TYPE       (-8)        /* wrong field type */
#define E_STDIN      (-9)        /* stdin used more than once */
#define E_OPTARGVAL (-10)        /* invalid argument value */
#define E_NOCLSATT  (-11)        /* no class attribute found */
#define E_PARSE     (-12)        /* parse error */ 
#define E_TAB_ATTS  (-13)        /* tables do not share attributes */
#define E_OATTNAME  (-14)        /* output attribute name conflict */
#define E_OATTTYPE  (-15)        /* invalid output attribute class */
#define E_ATTVAL    (-16)        /* invalid attribute value encountered */
#define E_TATTNAME  (-17)        /* target attribute name already in use */
#define E_HDRSPEC   (-18)        /* illegal position for header specification */
#define E_DISTFUN   (-19)        /* unknown distance function name */
#define E_TCOLADD   (-20)        /* could not add table column */
#define E_CATTYPE   (-21)        /* non-nominal class attribute selected: */
#define E_UNKNOWN   (-22)        /* unknown error */
                  
/*------------------------------------------------------------------------------
  Constants 
------------------------------------------------------------------------------*/

static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error\n",
  /* E_NOMEM     1 */  "not enough memory\n",
  /* E_FOPEN     2 */  "cannot open file %s\n",
  /* E_FREAD     3 */  "read error on file %s\n",
  /* E_FWRITE    4 */  "write error on file %s\n",
  /* E_OPTION    5 */  "unknown option -%c\n",
  /* E_OPTARG    6 */  "missing option argument\n",
  /* E_ARGCNT    7 */  "wrong number of arguments\n\tuse -? option for help",
  /* E_TYPE      8 */  "wrong type (field must be numeric)\n",
  /* E_STDIN     9 */  "only one file may be assigned to stdin\n",
  /* E_OPTARGVAL 10 */ "invalid parameter value for %s\n", 
  /* E_NOCLSATT  11 */ "no class attribute found",
  /* E_PARSE     12 */ "parse error(s) on file %s\n",
  /* E_TAB_ATTS  13 */ "tables do not share attributes\n", 
  /* E_OATTNAME  14 */ "output attribute name used already\n", 
  /* E_OATTTYPE  15 */ "invalid output attribute type (attribute %s)\n",
  /* E_ATTVAL    16 */ "encountered invalid attribute value\n", 
  /* E_TATTNAME  17 */ "name for target attribute already used: %s\n",
  /* E_HDRSPEC   18 */ "illegal position for header specification\n",
  /* E_DISTFUN   19 */ "unknown distance function name: %s\n",
  /* E_TCOLADD   20 */ "could not add table column: %s - column name \
conflict?\n",
  /* E_CATTYPE   21 */ " non-nominal class attribute selected: %s \n",
  /* E_UNKNOWN   22 */ "unknown error\n"
};      

/* structure with column mapping and target attribute information 
   negative values in mapsto entries indicate that column is ignored
   otherwise the number of the target column is provided */
typedef struct _colmap {               /* --- column mapping ---     */
  int      dstcol;                     /* column no. for distance */
  int      mfrqcol;                    /* column no. for maximal class frq */
  int      outcol;                     /* output column number in data file */
  int      nnghcol;                    /* column for number of neighbors */
  int      clscol;                     /* for target attrib col. example file */
  int numcols;                         /* column number in 1st table */
  int *mapsto;                         /* column index in 2nd table  */
} COLMAP;

COLMAP* cm_create(TABLE* from, TABLE* to, const char* tcname);
void cm_delete(COLMAP* cm);


typedef struct _nghrec{               /* record neighbor search */
  INST   class;                       /* class for neighbor  */
  double dist;                        /* distance of neighbor */
} NGHREC;


/* a tuple distance function */
typedef double TPLDIST (const TUPLE*, const TUPLE*,const COLMAP*); 

#define cm_tcolidx(c)        ((c)->tcidx)
#define cm_mapidx(c,i)       ((c)->mapsto[i])


#ifndef NDEBUG
void cm_show(COLMAP* cm);
void as_show(ATTSET* attset);
#endif
 
#endif
